/*
 * Copyright (C) 2003 Momchil Velikov
 *
 * See file COPYING for license.
 */


#ifndef nonstd__signal_h
#define nonstd__signal_h 1

#include <iostream>
#include <vector>
#include <algorithm>
#include <functional>

#include "function.h"
#include "ptr-to-member.h"

namespace nonstd
{
  namespace impl
  {
    template<typename Tp> class signal_base;
  }

  template<typename Tp>
  class slot : public function<Tp>
  {
  public:
    template<typename Fn>
    slot (const Fn &fn) : function<Tp> (fn)
    {
    }

    template<typename Obj>
    slot (Obj *obj, typename ptr_to_member<Obj, Tp>::type fn)
      : function<Tp> (obj, fn)
    {
    }

    ~slot ()
    {
      detach ();
    }

    void
    detach ()
    {
      std::for_each (peers_.begin (), peers_.end (),
                     std::bind2nd (do_detach (), this));
      peers_.clear ();
    }

  protected:
    typedef class impl::signal_base<Tp> signal_type;
    friend class impl::signal_base<Tp>;

    void
    attach (signal_type *sig)
    {
      peers_.push_back (sig);
    }

    void
    detach (signal_type *sig)
    {
      typename peer_vec::iterator i;

      i = std::remove_if (peers_.begin (), peers_.end (),
                          std::bind2nd (std::equal_to<signal_type *> (),
                                        sig));
      peers_.erase (i, peers_.end ());
    }

    struct do_detach
      : public std::binary_function<signal_type *, slot<Tp> *, void>
    {
      void
      operator() (signal_type *sig, slot<Tp> *sl) const
      {
        sig->disconnect (sl);
      }
    };

    typedef std::vector<signal_type *> peer_vec;
    peer_vec peers_;
  };


  namespace impl
  {
    template<typename Tp>
    class signal_base
    {
    public:

      void
      connect (slot<Tp> *s)
      {
        s->attach (this);
        peers_.push_back (s);
      }

      void
      disconnect ()
      {
        std::for_each (peers_.begin (), peers_.end (),
                       std::bind2nd (do_disconnect (), this));
        peers_.clear ();
      }

      void
      disconnect (slot<Tp> *s)
      {
        typename peer_vec::iterator i;

        i = std::remove_if (peers_.begin (), peers_.end (),
                            std::bind2nd (std::equal_to<slot<Tp> *> (), s));
        peers_.erase (i, peers_.end ());
      }

    protected:
      struct do_disconnect
        : public std::binary_function<slot<Tp> *, signal_base<Tp> *, void>
      {
        void
        operator() (slot<Tp> *sl, signal_base<Tp> *sig) const
        {
          sl->detach (sig);
        }
      };

      typedef std::vector<slot<Tp> *> peer_vec;
      peer_vec peers_;
    };
  }


  namespace impl
  {
    template<unsigned int arity, typename Tp> struct signal_helper;

    template<typename Tp>
    struct signal_helper<0, Tp> : public signal_base<Tp>
    {
      void
      operator() ()
      {
        typename signal_base<Tp>::peer_vec::iterator i;

        for (i = this->peers_.begin (); i != this->peers_.end (); ++i)
          (**i) ();
      }
    };

    template<typename Tp>
    struct signal_helper<1, Tp> : public signal_base<Tp>
    {
      void
      operator() (typename function_traits<Tp>::arg0_type arg0)
      {
        typename signal_base<Tp>::peer_vec::iterator i;

        for (i = this->peers_.begin (); i != this->peers_.end (); ++i)
          (*i) (arg0);
      }
    };

    template<typename Tp>
    struct signal_helper<2, Tp> : public signal_base<Tp>
    {
      void
      operator() (typename function_traits<Tp>::arg0_type arg0,
                  typename function_traits<Tp>::arg1_type arg1)
      {
        typename signal_base<Tp>::peer_vec::iterator i;

        for (i = this->peers_.begin (); i != this->peers_.end (); ++i)
          (*i) (arg0, arg1);
      }
    };

    template<typename Tp>
    struct signal_helper<3, Tp> : public signal_base<Tp>
    {
      void
      operator() (typename function_traits<Tp>::arg0_type arg0,
                  typename function_traits<Tp>::arg1_type arg1,
                  typename function_traits<Tp>::arg2_type arg2)
      {
        typename signal_base<Tp>::peer_vec::iterator i;

        for (i = this->peers_.begin (); i != this->peers_.end (); ++i)
          (*i) (arg0, arg1, arg2);
      }
    };

    template<typename Tp>
    struct signal_helper<4, Tp> : public signal_base<Tp>
    {
      void
      operator() (typename function_traits<Tp>::arg0_type arg0,
                  typename function_traits<Tp>::arg1_type arg1,
                  typename function_traits<Tp>::arg2_type arg2,
                  typename function_traits<Tp>::arg3_type arg3)
      {
        typename signal_base<Tp>::peer_vec::iterator i;

        for (i = this->peers_.begin (); i != this->peers_.end (); ++i)
          (*i) (arg0, arg1, arg2, arg3);
      }
    };
  }


  template<typename Tp>
  struct signal : public impl::signal_helper<function_traits<Tp>::arity, Tp>
  {
  };
}

#endif /* nonstd__signal_h */

/*
 * Local variables:
 * mode: C++
 * indent-tabs-mode: nil
 * arch-tag: ebc872d9-3fa9-4169-8f33-f5fe23c2b7c2
 * End:
 */
