/*
 * Copyright (C) 2003 Momchil Velikov
 *
 * See file COPYING for license.
 */


#ifndef nonstd__token_h
#define nonstd__token_h 1

#include <iterator>
#include <string>

namespace nonstd
{
  template<typename Ch = char,
           typename Tp = std::basic_string<Ch>,
           typename It = typename std::basic_string<Ch>::const_iterator>
  class char_separator_fn
  {
  public:
    typedef Ch char_type;
    typedef Tp token_type;
    typedef It iterator_type;

    char_separator_fn (const Ch *drop_sep, const Ch *kept_sep = 0,
                       const Ch *escape = 0, bool output_empty = false)
      : drop_sep_ (drop_sep), kept_sep_ (kept_sep), escape_ (escape),
        output_empty_ (output_empty), do_empty_ (true)
    {
    }

    bool
    operator() (It &begin, It end, Tp &token)
    {
      token = Tp ();

      /* Skip dropped separators, or output the empty tokens between
         them.  */
      while (begin != end && is_drop_separator (begin))
        {
          if (output_empty_ && do_empty_)
            {
              do_empty_ = false;
              return true;
            }
          else
            {
              do_empty_ = true;
              ++begin;
            }
        }

      if (begin == end)
        {
          if (output_empty_ && do_empty_)
            {
              do_empty_ = false;
              return true;
            }
          else
            return false;
        }

      /* If kept separator found, either output an empty token or the
         separator itself.  */
      if (is_kept_separator (begin))
        {
          if (output_empty_ && do_empty_)
            {
              do_empty_ = false;
              return true;
            }
          else
            {
              do_empty_ = true;
              token += *begin;
              ++begin;
              return true;
            }
        }

      /* Discard escape character.  */
      if (is_escape (begin))
        ++begin;

      if (begin == end)
        return do_empty_;

      /* Escaped character is always a part of the token.  */
      token += *begin;
      ++begin;

      while (begin != end
             && !is_kept_separator (begin)
             && !is_drop_separator (begin))
        {
          if (is_escape (begin))
            ++begin;

          if (begin == end)
            break;

          token += *begin;
          ++begin;
        }

      do_empty_ = false;
      return true;
    }

  protected:
    const Ch *drop_sep_;
    const Ch *kept_sep_;
    const Ch *escape_;
    bool output_empty_;
    bool do_empty_;

    typedef std::char_traits<Ch> traits;

    bool
    is_kept_separator (It in) const
    {
      return (kept_sep_ != 0
              && traits::find (kept_sep_, traits::length (kept_sep_), *in));
    }

    bool
    is_drop_separator (It in) const
    {
      return (drop_sep_ != 0
              && traits::find (drop_sep_, traits::length (drop_sep_), *in));
    }

    bool
    is_escape (It in)
    {
      return (escape_ != 0
              && traits::find (escape_, traits::length (escape_), *in));
    }
  };


  namespace impl
  {
    template<class Fn,
             typename Tp = typename Fn::token_type,
             typename It = typename Fn::iterator_type>
    class tokenizer_iterator
      : public std::iterator<std::input_iterator_tag, Tp, void>
    {
    public:
      tokenizer_iterator ()
      {
        has_token_ = false;
      }

      tokenizer_iterator (It end)
        : begin_ (end), end_ (end)
      {
        has_token_ = false;
      }

      tokenizer_iterator (It begin, It end)
        : begin_ (begin), end_ (end)
      {
        has_token_ = func_ (begin_, end_, token_);
      }

      tokenizer_iterator (const tokenizer_iterator<Fn, Tp, It> &i)
        : func_ (i.func_), token_ (i.token_), begin_ (i.begin_), end_ (i.end_),
          has_token_ (i.has_token_)
      {
      }

      tokenizer_iterator<Fn, Tp, It> &
      operator= (const tokenizer_iterator<Fn, Tp, It> &i)
      {
        func_ = i.func_;
        token_ = i.token_;
        begin_ = i.begin_;
        end_ = i.end_;
        has_token_ = i.has_token_;

        return *this;
      }

      bool
      operator== (const tokenizer_iterator<Fn, Tp, It> &i) const
      {
        return (has_token_ == i.has_token_ && begin_ == i.begin_
                && end_ == i.end_);
      }

      bool
      operator!= (const tokenizer_iterator<Fn, Tp, It> &i) const
      {
        return (has_token_ != i.has_token_ || begin_ != i.begin_
                || end_ != i.end_);
      }

      Tp &
      operator* ()
      {
        return token_;
      }

      const Tp &
      operator* () const
      {
        return token_;
      }

      Tp *
      operator-> ()
      {
        return &token_;
      }

      const Tp *
      operator-> () const
      {
        return &token_;
      }

      tokenizer_iterator<Fn, Tp, It> &
      operator++ ()
      {
        has_token_ = func_ (begin_, end_, token_);
        return *this;
      }

      tokenizer_iterator<Fn, Tp, It>
      operator++ (int)
      {
        tokenizer_iterator<Fn, Tp, It> it (*this);
        has_token_ = func_ (begin_, end_, token_);
        return it;
      }

      Fn func_;
      Tp token_;
      It begin_;
      It end_;
      bool has_token_;
    };
  }


  template<class Fn,
           typename Tp = typename Fn::token_type,
           typename It = typename Fn::iterator_type>
  class tokenizer
  {
  public:
    typedef impl::tokenizer_iterator<Fn, Tp, It> iterator;

    tokenizer (It begin, It end)
      : begin_ (begin, end)
    {
    }

    template<typename C>
    tokenizer (const C &c)
      : begin_ (c.begin (), c.end ())
    {
    }

    iterator
    begin () const
    {
      return begin_;
    }

    iterator
    end () const
    {
      return iterator (begin_.end_);
    }

  protected:
    iterator begin_;
  };
}

#endif /* nonstd__token_h */

/*
 * Local variables:
 * mode: C++
 * indent-tabs-mode: nil
 * arch-tag: 464a4bd6-bbda-4d8d-8457-295bd571d7b1
 * End:
 */
